# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit
from jax import numpy as jnp
from jax.scipy.stats import truncnorm as jax_truncnorm
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class TruncNormal(ContinuousRV):

    def __init__(self,
                 mu: ArrayLike,
                 sigma: ArrayLike,
                 low: ArrayLike = 0.0,
                 high: ArrayLike = 1.0,
                 name: str = None) -> None:
        self._mu = mu
        self._sigma = sigma
        self._low = low
        self._high = high
        self.check_params()
        self._alpha = (self._low - self._mu) / self._sigma
        self._beta = (self._high - self._mu) / self._sigma
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._low < self._high), "low must be smaller than high"
        assert jnp.all(self._sigma > 0), "sigma must be positive"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        return jax_truncnorm.logpdf(x, self._alpha, self._beta, loc=self._mu, scale=self._sigma)

    @partial(jit, static_argnums=(0,))
    def pdf(self, x: ArrayLike) -> ArrayLike:
        return jax_truncnorm.pdf(x, self._alpha, self._beta, loc=self._mu, scale=self._sigma)

    @partial(jit, static_argnums=(0,))
    def logcdf(self, x: ArrayLike) -> ArrayLike:
        return jax_truncnorm.logcdf(x, self._alpha, self._beta, loc=self._mu, scale=self._sigma)

    @partial(jit, static_argnums=(0,))
    def cdf(self, x: ArrayLike) -> ArrayLike:
        return jax_truncnorm.cdf(x, self._alpha, self._beta, loc=self._mu, scale=self._sigma)

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        return jax.random.truncated_normal(key, self._alpha, self._beta, shape=(N,)) * self._sigma + self._mu

    def __repr__(self) -> str:
        string = f"TruncNorm(mu={self._mu}, sigma={self._sigma}, low={self._low}, high={self._high}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit
from jax import numpy as jnp
from jax.scipy.stats import geom as jax_geom
from jax.typing import ArrayLike

from .drvs import DiscreteRV


class Geometric(DiscreteRV):
    """Geometric Random Variable"""

    def __init__(self, p: ArrayLike, name: str = None) -> None:
        self._p = p
        self.check_params()
        self._q = 1.0 - self._p
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._p >= 0.0), "All p must be greater than or equals to 0"

    @partial(jit, static_argnums=(0,))
    def logpmf(self, k: ArrayLike) -> ArrayLike:
        return jax_geom.logpmf(k, self._p)

    @partial(jit, static_argnums=(0,))
    def pmf(self, k: ArrayLike) -> ArrayLike:
        return jax_geom.pmf(k, self._p)

    @partial(jit, static_argnums=(0,))
    def cdf(self, k: ArrayLike) -> ArrayLike:
        conditions = [k < 0, k >= 0]
        choices = [jnp.zeros_like(k), 1.0 - jnp.power(self._q, jnp.floor(k))]
        return jnp.select(conditions, choices)

    @partial(jit, static_argnums=(0,))
    def logcdf(self, k: ArrayLike) -> ArrayLike:
        return jnp.log(self.cdf(k))

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        return jax.random.geometric(key, self._p, shape=(N,))

    def __repr__(self) -> str:
        string = f"Geometric(p={self._p}"
        if self._name is not None:
            string += f", (name: {self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax.typing import ArrayLike

from .binomial import Binomial


class Bernoulli(Binomial):
    """Bernoulli random variable"""

    def __init__(self, p: ArrayLike, name: str = None) -> None:
        """Initialize the Bernoulli random variable.

        Parameters
        ----------
        p : ArrayLike
            Probability of success.
        name : str, optional
            Name of the random variable, by default None
        """
        super().__init__(p, 1, name)

    def __repr__(self) -> str:
        """Return the string representation of the Bernoulli random variable.

        Returns
        -------
        str
            String representation of the Bernoulli random variable.
        """
        string = f"Bernoulli(p={self._p}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .beta import Beta
from .boltzmann import Boltzmann
from .cauchy import Cauchy
from .chi2 import Chi2
from .crvs import ContinuousRV
from .exponential import Exponential
from .gamma import Gamma
from .logistic import Logistic
from .lognormal import LogNormal
from .normal import Normal
from .pareto import Pareto
from .rayleigh import Rayleigh
from .studentt import StudentT
from .triangular import Triangular
from .truncnormal import TruncNormal
from .truncpowerlaw import TruncPowerLaw
from .uniform import Uniform
from .weibull import Weibull



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import Array
from matplotlib import pyplot as plt

from ..rvs import ContinuousRV
from .sampler import Sampler


class InverseTransformSampler(Sampler):

    def __init__(self) -> None:
        super().__init__()

    def sample(self, rv: ContinuousRV, N: int = 1, key: Array = None, scatter_plot: bool = False) -> Array:
        self.check_rv(rv)

        if key is None:
            key = self.get_key(key)

        U = jax.random.uniform(key, shape=(N,))
        samples = rv.ppf(U)

        if scatter_plot:
            plt.scatter(
                U,
                rv.cdf(U),
                c=samples,
                cmap="viridis",
                alpha=0.5,
                label="Inverse Transform samples",
            )
        return samples



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit, lax
from jax import numpy as jnp
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class TruncPowerLaw(ContinuousRV):

    def __init__(self, alpha: ArrayLike, low: ArrayLike = 0, high: ArrayLike = 1, name: str = None) -> None:
        self._alpha = alpha
        self._low = low
        self._high = high
        self.check_params()
        self._beta = 1.0 + self._alpha
        self._logZ = self.logZ()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._low > 0.0), "low must be greater than 0"
        assert jnp.all(self._high > self._low), "high must be greater than low"

    @partial(jit, static_argnums=(0,))
    def logZ(self) -> ArrayLike:
        logZ_val = lax.cond(
            self._alpha == -1.0, lambda _: jnp.log(jnp.log(self._high) - jnp.log(self._low)), lambda beta: lax.cond(
                beta > 0.0,
                lambda b: jnp.log(jnp.power(self._high, b) - jnp.power(self._low, b)) - jnp.log(b),
                lambda b: jnp.log(jnp.power(self._low, b) - jnp.power(self._high, b)) - jnp.log(-b),
                self._beta,
            ), self._beta)
        return logZ_val

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        logpdf_val = jnp.log(x) * self._alpha - self._logZ
        logpdf_val = jnp.where((x >= self._low) * (x <= self._high), logpdf_val, -jnp.inf)
        return logpdf_val

    @partial(jit, static_argnums=(0,))
    def logcdf(self, x: ArrayLike) -> ArrayLike:
        logcdf_val = lax.cond(
            self._alpha == -1.0, lambda _: jnp.log(jnp.log(x) - jnp.log(self._low)), lambda beta: lax.cond(
                beta > 0.0,
                lambda b: jnp.log(jnp.power(x, b) - jnp.power(self._low, b)) - jnp.log(b),
                lambda b: jnp.log(jnp.power(self._low, b) - jnp.power(x, b)) - jnp.log(-b),
                self._beta,
            ), self._beta)
        logcdf_val -= self._logZ
        logcdf_val = jnp.where(x >= self._low, logcdf_val, -jnp.inf)
        logcdf_val = jnp.where(x <= self._high, logcdf_val, jnp.log(1.0))
        return logcdf_val

    @partial(jit, static_argnums=(0,))
    def logppf(self, x: ArrayLike) -> ArrayLike:
        logppf_val = lax.cond(
            self._alpha == -1.0,
            lambda _: x * jnp.log(self._high) + (1.0 - x) * jnp.log(self._low),
            lambda b: (1.0 / b) * jnp.log(x * b * jnp.exp(self._logZ) + jnp.power(self._low, b)),
            self._beta,
        )
        logppf_val = jnp.where(x >= 0.0, logppf_val, -jnp.inf)
        logppf_val = jnp.where(x <= 1.0, logppf_val, jnp.log(1.0))
        return logppf_val

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        U = jax.random.uniform(key, shape=(N,), dtype=jnp.float32)
        rvs_val = self.ppf(U)
        return rvs_val

    def __repr__(self) -> str:
        string = f"TruncPowerLaw(alpha={self._alpha}, low={self._low}, high={self._high}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit, lax
from jax import numpy as jnp
from jax.scipy.stats import uniform as jax_uniform
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class Uniform(ContinuousRV):

    def __init__(self, low: ArrayLike = 0.0, high: ArrayLike = 1.0, name: str = None) -> None:
        self._low = low
        self._high = high
        self.check_params()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._low < self._high), "All low must be less than high"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        return jax_uniform.logpdf(x, loc=self._low, scale=self._high - self._low)

    @partial(jit, static_argnums=(0,))
    def pdf(self, x: ArrayLike) -> ArrayLike:
        return jax_uniform.pdf(x, loc=self._low, scale=self._high - self._low)

    @partial(jit, static_argnums=(0,))
    def logcdf(self, x: ArrayLike) -> ArrayLike:
        conditions = [x < self._low, (self._low <= x) & (x <= self._high), self._high < x]
        choice = [
            -jnp.inf,
            lax.log(x - self._low) - lax.log(self._high - self._low),
            jnp.log(1.0),
        ]
        return jnp.select(conditions, choice)

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        return jax.random.uniform(key, minval=self._low, maxval=self._high, shape=(N,))

    def __repr__(self) -> str:
        string = f"Uniform(low={self._low}, high={self._high}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import Array
from jax import numpy as jnp
from matplotlib import pyplot as plt

from ..rvs import ContinuousRV
from .sampler import Sampler


class AdaptiveAcceptRejectSampler(Sampler):

    def __init__(self) -> None:
        super().__init__()

    def sample(self,
               target_rv: ContinuousRV,
               proposal_rv: ContinuousRV,
               scale: int = 1,
               N: int = 1,
               key: Array = None,
               scatter_plot: bool = False) -> Array:
        self.check_rv(target_rv)
        self.check_rv(proposal_rv)
        if key is None:
            key = self.get_key(key)
        samples = jnp.array([])
        N_res = N
        while N_res != 0:
            N_res = N - len(samples)

            V = proposal_rv.rvs(N_res, key)
            key = self.get_key(key)
            pdf_ratio = target_rv.pdf(V)

            U_scaled = jax.random.uniform(
                key,
                shape=(N_res,),
                minval=0.0,
                maxval=scale * proposal_rv.pdf(V),
            )

            accept = U_scaled <= pdf_ratio
            samples = jnp.concatenate([samples, V[accept]])

            key = self.get_key(key)

        if scatter_plot:
            plt.scatter(
                V,
                U_scaled,
                c=accept,
                cmap="viridis",
                alpha=0.5,
                label="Accept/Reject samples",
            )
        return samples



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit, lax
from jax import numpy as jnp
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class Triangular(ContinuousRV):

    def __init__(self, low: float = 0, mode: float = 0.5, high: float = 1, name: str = None) -> None:
        self._low = low
        self._mode = mode
        self._high = high
        self.check_params()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._low <= self._high), "low must be less than or equal to high"
        assert jnp.all(self._low <= self._mode), "low must be less than or equal to mid"
        assert jnp.all(self._mode <= self._high), "mid must be less than or equal to high"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        conditions = [
            x < self._low,
            (self._low <= x) & (x < self._mode),
            x == self._mode,
            (self._mode < x) & (x <= self._high),
            x > self._high,
        ]
        choices = [
            -jnp.inf,
            jnp.log(2) + jnp.log(x - self._low) - jnp.log(self._high - self._low) - jnp.log(self._mode - self._low),
            jnp.log(2) - jnp.log(self._high - self._low),
            jnp.log(2) + jnp.log(self._high - x) - jnp.log(self._high - self._low) - jnp.log(self._high - self._mode),
            -jnp.inf,
        ]
        return jnp.select(conditions, choices)

    @partial(jit, static_argnums=(0,))
    def logcdf(self, x: ArrayLike) -> ArrayLike:
        conditions = [
            x < self._low,
            (self._low <= x) & (x < self._mode),
            x == self._mode,
            (self._mode < x) & (x < self._high),
            x >= self._high,
        ]
        choices = [
            -jnp.inf,
            2 * jnp.log(x - self._low) - jnp.log(self._high - self._low) - jnp.log(self._mode - self._low),
            jnp.log(0.5),
            jnp.log(1 - ((self._high - x)**2 / ((self._high - self._low) * (self._high - self._mode)))),
            jnp.log(1),
        ]
        return jnp.select(conditions, choices)

    @partial(jit, static_argnums=(0,))
    def ppf(self, x: ArrayLike) -> ArrayLike:
        _Fc = self.cdf(self._mode)
        ppf_val = jnp.where(
            x < _Fc,
            self._low + lax.sqrt(x * (self._mode - self._low) * (self._high - self._low)),
            self._high - lax.sqrt((1 - x) * (self._high - self._low) * (self._high - self._mode)),
        )
        return ppf_val

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        return jax.random.triangular(key, left=self._low, right=self._high, mode=self._mode, shape=(N,))

    def __repr__(self) -> str:
        string = f"Triangular(low={self._low}, mode={self._mode}, high={self._high}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit
from jax import numpy as jnp
from jax.scipy.stats import binom as jax_binom
from jax.typing import ArrayLike

from .drvs import DiscreteRV


class Binomial(DiscreteRV):
    """Binomial random variable"""

    def __init__(self, p: ArrayLike, n: int, name: str = None) -> None:
        """Initialize the Binomial random variable.

        Parameters
        ----------
        p : ArrayLike
            Probability of success.
        n : int
            Number of trials.
        name : str, optional
            Name of the random variable, by default None
        """
        self._p = p
        self._n = n
        self.check_params()
        self._q = 1.0 - p
        super().__init__(name)

    def check_params(self) -> None:
        """Check the parameters of the random variable."""
        assert jnp.all(self._p >= 0.0) and jnp.all(self._p <= 1.0), "p must be in [0, 1]"
        assert type(self._n) == int, "n must be an integer"
        assert self._n > 0, "n must be positive"

    @partial(jit, static_argnums=(0))
    def logpmf(self, k: ArrayLike) -> ArrayLike:
        k = jnp.asarray(k)
        return jax_binom.logpmf(k, self._n, self._p)

    @partial(jit, static_argnums=(0,))
    def pmf(self, k: ArrayLike) -> ArrayLike:
        k = jnp.asarray(k)
        return jax_binom.pmf(k, self._n, self._p)

    @partial(jit, static_argnums=(0,))
    def logcdf(self, k: ArrayLike) -> ArrayLike:
        return jnp.log(self.cdf(k))

    @partial(jit, static_argnums=(0,))
    def cdf(self, k: ArrayLike) -> ArrayLike:
        """Cumulative distribution function.

        Parameters
        ----------
        k : ArrayLike
            Input values.

        Returns
        -------
        ArrayLike
            Cumulative distribution function evaluated at k.
        """
        k = jnp.asarray(k)
        x = jnp.arange(0, self._n + 1, dtype=jnp.int32)
        complete_cdf = jnp.cumsum(self.pmf(x))
        cond = [k < 0, k >= self._n, jnp.logical_and(k >= 0, k < self._n)]
        return jnp.select(cond, [0.0, 1.0, complete_cdf[k]])

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        return jax.random.binomial(key=key, n=self._n, p=self._p, shape=(N,))

    def __repr__(self) -> str:
        string = f"Binomial(p={self._p}, n={self._n}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit
from jax import numpy as jnp
from jax.scipy.special import log_ndtr, ndtr, ndtri
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class LogNormal(ContinuousRV):

    def __init__(self, mu: ArrayLike = 0.0, sigma: ArrayLike = 1.0, name: str = None) -> None:
        self._mu = mu
        self._sigma = sigma
        self.check_params()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._sigma > 0.0), "All sigma must be greater than 0.0"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        constants = -(jnp.log(self._sigma) + 0.5 * jnp.log(2 * jnp.pi))
        logpdf_val = jnp.where(
            x <= 0,
            -jnp.inf,
            constants - jnp.log(x) - (0.5 * jnp.power(self._sigma, -2)) * jnp.power((jnp.log(x) - self._mu), 2),
        )
        return logpdf_val

    @partial(jit, static_argnums=(0,))
    def logcdf(self, x: ArrayLike) -> ArrayLike:
        return log_ndtr((jnp.log(x) - self._mu) / self._sigma)

    @partial(jit, static_argnums=(0,))
    def cdf(self, x: ArrayLike) -> ArrayLike:
        return ndtr((jnp.log(x) - self._mu) / self._sigma)

    @partial(jit, static_argnums=(0,))
    def ppf(self, x: ArrayLike) -> ArrayLike:
        return jnp.exp(self._mu + self._sigma * ndtri(x))

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        # return jax.random.lognormal(self.get_key(), shape=(N,)) * self._sigma + self._mu
        if key is None:
            key = self.get_key(key)
        U = jax.random.uniform(key, shape=(N,))
        return self.ppf(U)

    def __repr__(self) -> str:
        string = f"LogNormal(mu={self._mu}, sigma={self._sigma}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from jax import jit
from jax import numpy as jnp
from jax.typing import ArrayLike

from ..rvs import GenericRV


class DiscreteRV(GenericRV):

    def __init__(self, name: str = None) -> None:
        super().__init__(name)

    @partial(jit, static_argnums=(0,))
    def logpmf(self, *k: ArrayLike) -> ArrayLike:
        """Logarithm of the probability mass function.

        Parameters
        ----------
        *k : ArrayLike
            Input values.   

        Returns
        -------
        ArrayLike
            Logarithm of the probability mass function evaluated at *k.

        Raises
        ------
        NotImplementedError
            If the child class has not implemented this method.
        """
        raise NotImplementedError

    @partial(jit, static_argnums=(0,))
    def pmf(self, *k: ArrayLike) -> ArrayLike:
        """Probability mass function.

        Parameters
        ----------
        *k : ArrayLike
            Input values.

        Returns
        -------
        ArrayLike
            Probability mass function evaluated at *k.
        """
        return jnp.exp(self.logpmf(*k))

    def __str__(self) -> str:
        return super().__str__()



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random

import jax
from jax import Array

fact = [1, 1, 2, 6, 24, 120, 720, 5_040, 40_320, 362_880, 3_628_800]


def nPr(n: int, r: int) -> int:
    """Calculates the number of permutations of `r` objects out of `n`

    Parameters
    ----------
    n : int
        total objects
    r : int
        selected objects

    Returns
    -------
    int
        number of permutations of `r` objects out of `n`
    """
    assert 0 <= r <= n
    if n <= len(fact):
        return fact[n] / fact[r] / fact[n - r]
    for i in range(len(fact), n + 1):
        fact.append(fact[i - 1] * i)
    return fact[n] / fact[n - r]


def nCr(n: int, r: int) -> int:
    """Calculates the number of combinations of `r` objects out of `n`

    Parameters
    ----------
    n : int
        total objects
    r : int
        selected objects

    Returns
    -------
    int
        number of combinations of `r` objects out of `n`
    """
    assert 0 <= r <= n
    if n <= len(fact):
        return fact[n] / fact[r] / fact[n - r]
    for i in range(len(fact), n + 1):
        fact.append(fact[i - 1] * i)
    return fact[n] / fact[r] / fact[n - r]


def new_prn_key(key: Array = None) -> Array:
    """Get a new JAX random key.

    This function is used to generate a new JAX random key if
    the user does not provide one. The key is generated using
    the JAX random.PRNGKey function. The key is split into
    two keys, the first of which is returned. The second key
    is discarded.

    Parameters
    ----------
    key : Array, optional
        JAX random key, by default None

    Returns
    -------
    Array
        New JAX random key.
    """
    if key is None:
        new_key = jax.random.PRNGKey(random.randint(0, 1000_000))
    else:
        new_key, _ = jax.random.split(key)
    return new_key



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

from jax import Array

from ..rvs import ContinuousRV, GenericRV
from ..utils import new_prn_key


class Sampler(object):

    def __init__(self) -> None:
        pass

    def check_rv(self, rv: GenericRV) -> None:
        assert isinstance(rv, GenericRV), f"rv must be a GenericRV object, got {rv}"
        assert isinstance(rv, ContinuousRV), f"rv must be a ContinuousRV object, got {rv}"

    @abstractmethod
    def sample(self, rv: GenericRV, N: int = 1, key: Array = None) -> Array:
        raise NotImplementedError

    @staticmethod
    def get_key(key: Array = None) -> Array:
        return new_prn_key(key)



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import Array
from matplotlib import pyplot as plt

from ..rvs import ContinuousRV
from .sampler import Sampler


class AcceptRejectSampler(Sampler):

    def __init__(self) -> None:
        super().__init__()

    def sample(self,
               target_rv: ContinuousRV,
               proposal_rv: ContinuousRV,
               scale: int = 1,
               N: int = 1,
               key: Array = None,
               scatter_plot: bool = False) -> Array:
        self.check_rv(target_rv)
        self.check_rv(proposal_rv)

        if key is None:
            key = self.get_key(key)

        V = proposal_rv.rvs(N, key)

        pdf_ratio = target_rv.pdf(V)

        key = self.get_key(key)
        U_scaled = jax.random.uniform(
            key,
            shape=(N,),
            minval=0.0,
            maxval=scale * proposal_rv.pdf(V),
        )

        accept = U_scaled <= pdf_ratio
        samples = V[accept]

        if scatter_plot:
            plt.scatter(
                V,
                U_scaled,
                c=accept,
                cmap="viridis",
                alpha=0.5,
                label="Accept/Reject samples",
            )
        return samples



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax.random
from jax import Array, jit
from jax import numpy as jnp
from jax.typing import ArrayLike

from ..utils import new_prn_key


class GenericRV(object):
    """Generic random variable class."""

    def __init__(self, name: str = None) -> None:
        """Initialize the random variable.

        Parameters
        ----------
        name : str, optional
            Name of the random variable, by default None
        """
        self._name = name

    def check_params(self) -> None:
        """Check the parameters of the random variable.

        This method should be implemented by the child class.

        Raises
        ------
        NotImplementedError
            If the child class has not implemented this method.
        """
        raise NotImplementedError

    @partial(jit, static_argnums=(0,))
    def logcdf(self, *x: ArrayLike) -> ArrayLike:
        """Logarithm of the cumulative distribution function.
        
        Parameters
        ----------
        *x : ArrayLike
            Input values.

        Returns
        -------
        ArrayLike
            The logarithm of the cumulative distribution function.

        Raises
        ------
        NotImplementedError
            If the child class has not implemented this method.
        """
        raise NotImplementedError

    @partial(jit, static_argnums=(0,))
    def cdf(self, *x: ArrayLike) -> ArrayLike:
        """Cumulative distribution function.
        
        Parameters
        ----------
        *x : ArrayLike
            Input values.

        Returns
        -------
        ArrayLike
            The cumulative distribution function evaluated at x.
        """
        return jnp.exp(self.logcdf(*x))

    @partial(jit, static_argnums=(0,))
    def logppf(self, *x: ArrayLike) -> ArrayLike:
        """Logarithm of the percent point function.

        Returns
        -------
        ArrayLike
            The logarithm of the percent point function evaluated at x.

        Raises
        ------
        NotImplementedError
            If the child class has not implemented this method.
        """
        raise NotImplementedError

    @partial(jit, static_argnums=(0,))
    def ppf(self, *x: ArrayLike) -> ArrayLike:
        """Percent point function.

        Returns
        -------
        ArrayLike
            The percent point function evaluated at x.
        """
        return jnp.exp(self.logppf(*x))

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        """Random variates from the distribution of size N.

        Parameters
        ----------
        N : int, optional
            Number of random variates, by default 1
        key : Array, optional
            JAX random key, by default None

        Returns
        -------
        Array
            Random variates from the distribution of size N.

        Raises
        ------
        NotImplementedError
            If the child class has not implemented this method.
        """
        raise NotImplementedError

    @staticmethod
    def get_key(key: Array = None) -> Array:
        return new_prn_key(key)

    def __str__(self) -> str:
        """String representation of the random variable.

        Returns
        -------
        str
            String representation of the random variable.
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """String representation of the random variable.

        Returns
        -------
        str
            String representation of the random variable.

        Raises
        ------
        NotImplementedError
            If the child class has not implemented this method.
        """
        raise NotImplementedError



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from jax import jit
from jax import numpy as jnp
from jax.scipy.special import erf
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class Boltzmann(ContinuousRV):

    def __init__(self, a: ArrayLike, name: str = None) -> None:
        self._a = a
        self.check_params()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._a > 0.0), "a must be positive"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        logpdf_val = 2 * jnp.log(x) - 0.5 * jnp.power(x / self._a, 2)
        logpdf_val -= 0.5 * jnp.log(jnp.pi / 2) + 3 * jnp.log(self._a)
        logpdf_val = jnp.where(x > 0.0, logpdf_val, jnp.nan)
        return logpdf_val

    @partial(jit, static_argnums=(0,))
    def cdf(self, x: ArrayLike) -> ArrayLike:
        cdf_val = jnp.log(x) - 0.5 * jnp.power(x / self._a, 2)
        cdf_val -= 0.5 * jnp.log(jnp.pi / 2) + jnp.log(self._a)
        cdf_val = jnp.exp(cdf_val)
        cdf_val = erf(x / (jnp.sqrt(2) * self._a)) - cdf_val
        return cdf_val

    def __repr__(self) -> str:
        string = f"Boltzmann(a={self._a}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit
from jax import numpy as jnp
from jax.scipy.stats import beta as jax_beta
from jax.typing import ArrayLike
from tensorflow_probability.substrates import jax as tfp

from .crvs import ContinuousRV


class Beta(ContinuousRV):

    def __init__(self, alpha: ArrayLike, beta: ArrayLike, name: str = None) -> None:
        self._alpha = alpha
        self._beta = beta
        self.check_params()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._alpha > 0.0), "alpha must be positive"
        assert jnp.all(self._beta > 0.0), "beta must be positive"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        return jax_beta.logpdf(x, self._alpha, self._beta)

    @partial(jit, static_argnums=(0,))
    def pdf(self, x: ArrayLike) -> ArrayLike:
        return jax_beta.pdf(x, self._alpha, self._beta)

    @partial(jit, static_argnums=(0,))
    def logcdf(self, x: ArrayLike) -> ArrayLike:
        return jax_beta.logcdf(x, self._alpha, self._beta)

    @partial(jit, static_argnums=(0,))
    def cdf(self, x: ArrayLike) -> ArrayLike:
        return jax_beta.cdf(x, self._alpha, self._beta)

    @partial(jit, static_argnums=(0,))
    def ppf(self, x: ArrayLike) -> ArrayLike:
        return tfp.math.betaincinv(self._alpha, self._beta, x)

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        return jax.random.beta(key, self._alpha, self._beta, shape=(N,))

    def __repr__(self) -> str:
        string = f"Beta(alpha={self._alpha}, beta={self._beta}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_packages, setup

_current_version = '0.0.3'

with open('README.md', encoding='utf-8') as f:
    _long_description = f.read()

setup(
    name='jaxampler',
    version=_current_version,
    packages=find_packages(exclude=[
        'tests',
        'tests.*',
        'examples',
        'examples.*',
        'plots',
        'codesnap',
    ]),
    url='https://github.com/Qazalbash/jaxampler',
    license='Apache 2.0',
    author='Meesum Qazalbash',
    author_email='meesumqazalbash@gmail.com',
    description='JAX based lib for sampling statistical distributions.',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.9',
    install_requires=[
        'jaxlib>=0.4.0',
        'jax>=0.4.0',
        'tfp-nightly',
        'matplotlib>=3.8.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
    ],
    zip_safe=False,
)



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from jax import jit
from jax import numpy as jnp
from jax.scipy.special import betainc
from jax.scipy.stats import t as jax_t
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class StudentT(ContinuousRV):

    def __init__(self, nu: ArrayLike, name: str = None) -> None:
        self._nu = nu
        self.check_params()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._nu > 0.0), "nu must be positive"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        return jax_t.logpdf(x, self._nu)

    @partial(jit, static_argnums=(0,))
    def pdf(self, x: ArrayLike) -> ArrayLike:
        return jax_t.pdf(x, self._nu)

    @partial(jit, static_argnums=(0,))
    def cdf(self, x: ArrayLike) -> ArrayLike:
        return 1 - 0.5 * betainc(self._nu * 0.5, 0.5, 1 / (1 + jnp.power(x, 2) / self._nu))

    @partial(jit, static_argnums=(0,))
    def ppf(self, x: ArrayLike) -> ArrayLike:
        """A method is addressed in this paper https://www.homepages.ucl.ac.uk/~ucahwts/lgsnotes/JCF_Student.pdf"""
        raise NotImplementedError

    def __repr__(self) -> str:
        string = f"StudentT(nu={self._nu}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string



# Copyright 2023 The JAXampler Authors

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax
from jax import Array, jit
from jax import numpy as jnp
from jax.scipy.stats import pareto as jax_pareto
from jax.typing import ArrayLike

from .crvs import ContinuousRV


class Pareto(ContinuousRV):

    def __init__(self, alpha: ArrayLike, scale: ArrayLike, name: str = None) -> None:
        self._alpha = alpha
        self._scale = scale
        self.check_params()
        super().__init__(name)

    def check_params(self) -> None:
        assert jnp.all(self._alpha > 0.0), "alpha must be greater than 0"
        assert jnp.all(self._scale > 0.0), "scale must be greater than 0"

    @partial(jit, static_argnums=(0,))
    def logpdf(self, x: ArrayLike) -> ArrayLike:
        return jax_pareto.logpdf(x, self._alpha, scale=self._scale)

    @partial(jit, static_argnums=(0,))
    def pdf(self, x: ArrayLike) -> ArrayLike:
        return jax_pareto.pdf(x, self._alpha, scale=self._scale)

    @partial(jit, static_argnums=(0,))
    def logcdf(self, x: ArrayLike) -> ArrayLike:
        return jnp.where(self._scale <= x, jnp.log1p(-jnp.power(self._scale / x, self._alpha)), -jnp.inf)

    @partial(jit, static_argnums=(0,))
    def logppf(self, x: ArrayLike) -> ArrayLike:
        conditions = [
            x < 0.0,
            (0.0 <= x) & (x < 1.0),
            1.0 <= x,
        ]
        choices = [
            -jnp.inf,
            jnp.log(self._scale) - (1.0 / self._alpha) * jnp.log1p(-x),
            jnp.log(1.0),
        ]
        return jnp.select(conditions, choices)

    def rvs(self, N: int = 1, key: Array = None) -> Array:
        if key is None:
            key = self.get_key(key)
        return jax.random.pareto(key, self._alpha, shape=(N,)) * self._scale

    def __repr__(self) -> str:
        string = f"Pareto(alpha={self._alpha}, scale={self._scale}"
        if self._name is not None:
            string += f", name={self._name}"
        string += ")"
        return string
